"""Import utils."""
import importlib.metadata
import importlib.util
import logging
import sys
from pathlib import Path

logger = logging.getLogger(__name__)

PANDAS_IMPORT_ERROR = """
`{0}` requires the pandas library but it was not found in your environment. Please install fondant
 using the 'pandas' extra.
"""

DATASETS_IMPORT_ERROR = """
`{0}` requires the 🤗 Datasets library but it was not found in your environment.
Please install fondant using the 'datasets' extra.
Note that if you have a local folder named `datasets` or a local python file named
 `datasets.py` in your current working directory, python may try to import this instead of the 🤗
 Datasets library. You should rename this folder or that python file if that's the case.
  Please note that you may need to restart your runtime after installation.
"""

KFP_IMPORT_ERROR = """
`{0}` requires the kubeflow pipelines (kfp) library but it was not found in your environment.
Please install fondant using the 'pipelines' extra.
"""


def is_package_available(package_name: str, import_error_msg: str) -> bool:
    """
    Function that checks if a given package is available
    Args:
        package_name (str): the name of the package
        import_error_msg (str): the error message to return if the package is not found
    Returns:
        bool: check if package is available.
    """
    package_available = importlib.util.find_spec(package_name) is not None

    try:
        package_version = importlib.metadata.version(package_name)
        logger.debug(f"Successfully imported {package_name} version {package_version}")
    except importlib.metadata.PackageNotFoundError:
        package_available = False

    if package_available:
        return package_available

    raise ModuleNotFoundError(import_error_msg.format(Path(sys.argv[0]).stem))


def is_datasets_available():
    """Check if 'datasets' is available."""
    return is_package_available("datasets", DATASETS_IMPORT_ERROR)


def is_pandas_available():
    """Check if 'pandas' is available."""
    return is_package_available("pandas", PANDAS_IMPORT_ERROR)


def is_kfp_available():
    """Check if 'pandas' is available."""
    return is_package_available("kfp", KFP_IMPORT_ERROR)



"""
This component estimates the code to comments ratio and filters instances between two chosen
minimum and maximum values.
"""
import logging

import dask.dataframe as dd
from utils.text_extraction import get_comments_to_code_ratio

from fondant.component import DaskTransformComponent

logger = logging.getLogger(__name__)


class FilterCommentsComponent(DaskTransformComponent):
    """Component that filters instances based on code to comments ratio."""

    def transform(
        self,
        *,
        dataframe: dd.DataFrame,
        min_comments_ratio: float,
        max_comments_ratio: float,
    ) -> dd.DataFrame:
        """
        Args:
            dataframe: Dask dataframe
            min_comments_ratio: The minimum code to comment ratio
            max_comments_ratio: The maximum code to comment ratio
        Returns:
            Filtered dask dataframe.
        """
        # Apply the function to the desired column and filter the DataFrame
        return dataframe[
            dataframe["code_content"].map_partitions(
                lambda example: example.map(get_comments_to_code_ratio).between(
                    min_comments_ratio, max_comments_ratio,
                ),
            )
        ]


if __name__ == "__main__":
    component = FilterCommentsComponent.from_args()
    component.run()



"""
This Python module defines abstract base class for components in the Fondant data processing
framework, providing a standardized interface for extending loading and transforming components.
The loading component is the first component that loads the initial dataset and the transform
components take care of processing, filtering and extending the data.
"""

import argparse
import json
import logging
import typing as t
from abc import ABC, abstractmethod
from pathlib import Path

import dask.dataframe as dd
import pandas as pd

from fondant.component_spec import Argument, ComponentSpec, kubeflow2python_type
from fondant.data_io import DaskDataLoader, DaskDataWriter
from fondant.manifest import Manifest

logger = logging.getLogger(__name__)


class Component(ABC):
    """Abstract base class for a Fondant component."""

    def __init__(
        self,
        spec: ComponentSpec,
        *,
        input_manifest_path: t.Union[str, Path],
        output_manifest_path: t.Union[str, Path],
        metadata: t.Dict[str, t.Any],
        user_arguments: t.Dict[str, Argument],
    ) -> None:
        self.spec = spec
        self.input_manifest_path = input_manifest_path
        self.output_manifest_path = output_manifest_path
        self.metadata = metadata
        self.user_arguments = user_arguments

    @classmethod
    def from_file(
        cls,
        path: t.Union[str, Path] = "../fondant_component.yaml",
    ) -> "Component":
        """Create a component from a component spec file.

        Args:
            path: Path to the component spec file
        """
        component_spec = ComponentSpec.from_file(path)
        return cls.from_spec(component_spec)

    @classmethod
    def from_args(cls) -> "Component":
        """Create a component from a passed argument containing the specification as a dict."""
        parser = argparse.ArgumentParser()
        parser.add_argument("--component_spec", type=json.loads)
        args, _ = parser.parse_known_args()

        if "component_spec" not in args:
            msg = "Error: The --component_spec argument is required."
            raise ValueError(msg)

        component_spec = ComponentSpec(args.component_spec)

        return cls.from_spec(component_spec)

    @classmethod
    def from_spec(cls, component_spec: ComponentSpec) -> "Component":
        """Create a component from a component spec."""
        args_dict = vars(cls._add_and_parse_args(component_spec))

        if "component_spec" in args_dict:
            args_dict.pop("component_spec")
        input_manifest_path = args_dict.pop("input_manifest_path")
        output_manifest_path = args_dict.pop("output_manifest_path")
        metadata = args_dict.pop("metadata")

        metadata = json.loads(metadata) if metadata else {}

        return cls(
            component_spec,
            input_manifest_path=input_manifest_path,
            output_manifest_path=output_manifest_path,
            metadata=metadata,
            user_arguments=args_dict,
        )

    @classmethod
    def _add_and_parse_args(cls, spec: ComponentSpec):
        parser = argparse.ArgumentParser()
        component_arguments = cls._get_component_arguments(spec)

        for arg in component_arguments.values():
            if arg.name in cls.optional_fondant_arguments():
                input_required = False
                default = None
            elif arg.default:
                input_required = False
                default = arg.default
            else:
                input_required = True
                default = None

            parser.add_argument(
                f"--{arg.name}",
                type=kubeflow2python_type(arg.type),  # type: ignore
                required=input_required,
                default=default,
                help=arg.description,
            )

        return parser.parse_args()

    @staticmethod
    def optional_fondant_arguments() -> t.List[str]:
        return []

    @staticmethod
    def _get_component_arguments(spec: ComponentSpec) -> t.Dict[str, Argument]:
        """
        Get the component arguments as a dictionary representation containing both input and output
            arguments of a component
        Args:
            spec: the component spec
        Returns:
            Input and output arguments of the component.
        """
        component_arguments: t.Dict[str, Argument] = {}
        kubeflow_component_spec = spec.kubeflow_specification
        component_arguments.update(kubeflow_component_spec.input_arguments)
        component_arguments.update(kubeflow_component_spec.output_arguments)
        return component_arguments

    @abstractmethod
    def _load_or_create_manifest(self) -> Manifest:
        """Abstract method that returns the dataset manifest."""

    @abstractmethod
    def _process_dataset(self, manifest: Manifest) -> t.Union[None, dd.DataFrame]:
        """Abstract method that processes the manifest and
        returns another dataframe.
        """

    def _write_data(self, dataframe: dd.DataFrame, *, manifest: Manifest):
        """Create a data writer given a manifest and writes out the index and subsets."""
        data_writer = DaskDataWriter(manifest=manifest, component_spec=self.spec)
        data_writer.write_dataframe(dataframe)

    def run(self):
        """Runs the component."""
        input_manifest = self._load_or_create_manifest()

        output_df = self._process_dataset(manifest=input_manifest)

        output_manifest = input_manifest.evolve(component_spec=self.spec)

        self._write_data(dataframe=output_df, manifest=output_manifest)

        self.upload_manifest(output_manifest, save_path=self.output_manifest_path)

    def upload_manifest(self, manifest: Manifest, save_path: str):
        Path(save_path).parent.mkdir(parents=True, exist_ok=True)
        manifest.to_file(save_path)


class LoadComponent(Component):
    """Base class for a Fondant load component."""

    @staticmethod
    def optional_fondant_arguments() -> t.List[str]:
        return ["input_manifest_path"]

    def _load_or_create_manifest(self) -> Manifest:
        component_id = self.spec.name.lower().replace(" ", "_")
        return Manifest.create(
            base_path=self.metadata["base_path"],
            run_id=self.metadata["run_id"],
            component_id=component_id,
        )

    @abstractmethod
    def load(self, *args, **kwargs) -> dd.DataFrame:
        """Abstract method that loads the initial dataframe."""

    def _process_dataset(self, manifest: Manifest) -> dd.DataFrame:
        """This function loads the initial dataframe sing the user-provided `load` method.

        Returns:
            A `dd.DataFrame` instance with initial data'.
        """
        # Load the dataframe according to the custom function provided to the user
        return self.load(**self.user_arguments)


class TransformComponent(Component):
    """Base class for a Fondant transform component."""

    def _load_or_create_manifest(self) -> Manifest:
        return Manifest.from_file(self.input_manifest_path)

    @abstractmethod
    def transform(self, *args, **kwargs) -> dd.DataFrame:
        """
        Abstract method for applying data transformations to the input dataframe.

        Args:
            args: The dataframe will be passed in as a positional argument
            kwargs: Arguments provided to the component are passed as keyword arguments
        """

    def _process_dataset(self, manifest: Manifest) -> dd.DataFrame:
        """
        Load the data based on the manifest using a DaskDataloader and call the transform method to
        process it.

        Returns:
            A `dd.DataFrame` instance with updated data based on the applied data transformations.
        """


class DaskTransformComponent(TransformComponent):
    @abstractmethod
    def transform(self, *args, **kwargs) -> dd.DataFrame:
        """
        Abstract method for applying data transformations to the input dataframe.

        Args:
            args: A Dask dataframe will be passed in as a positional argument
            kwargs: Arguments provided to the component are passed as keyword arguments
        """

    def _process_dataset(self, manifest: Manifest) -> dd.DataFrame:
        """
        Load the data based on the manifest using a DaskDataloader and call the transform method to
        process it.

        Returns:
            A `dd.DataFrame` instance with updated data based on the applied data transformations.
        """
        data_loader = DaskDataLoader(manifest=manifest, component_spec=self.spec)
        dataframe = data_loader.load_dataframe()
        dataframe = self.transform(dataframe, **self.user_arguments)
        return dataframe


class PandasTransformComponent(TransformComponent):
    def setup(self, *args, **kwargs):
        """Called once for each instance of the Component class. Use this to set up resources
        such as database connections.
        """
        return

    @abstractmethod
    def transform(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """
        Abstract method for applying data transformations to the input dataframe.
        Called once for each partition of the data.

        Args:
            dataframe: A Pandas dataframe containing a partition of the data
        """

    def wrapped_transform(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """Method wrapping the transform method to switch between hierarchical and flattened
        columns.
        """
        dataframe.columns = pd.MultiIndex.from_tuples(
            tuple(column.split("_")) for column in dataframe.columns
        )
        dataframe = self.transform(dataframe)
        # Drop columns not in the produces section of the component spec
        dataframe = dataframe.drop(
            columns=[
                (subset, field)
                for (subset, field) in dataframe.columns
                if subset not in self.spec.produces
                or field not in self.spec.produces[subset].fields
            ],
        )
        dataframe.columns = [
            "_".join(column) for column in dataframe.columns.to_flat_index()
        ]
        return dataframe

    def _process_dataset(self, manifest: Manifest) -> dd.DataFrame:
        """
        Load the data based on the manifest using a DaskDataloader and call the transform method to
        process it.

        Returns:
            A `dd.DataFrame` instance with updated data based on the applied data transformations.
        """
        data_loader = DaskDataLoader(manifest=manifest, component_spec=self.spec)
        dataframe = data_loader.load_dataframe()

        # Call the component setup method with user provided argument
        self.setup(**self.user_arguments)

        # Create meta dataframe with expected format
        meta_dict = {"id": pd.Series(dtype="object")}
        for subset_name, subset in self.spec.produces.items():
            for field_name, field in subset.fields.items():
                meta_dict[f"{subset_name}_{field_name}"] = pd.Series(
                    dtype=pd.ArrowDtype(field.type.value),
                )
        meta_df = pd.DataFrame(meta_dict).set_index("id")

        # Call the component transform method for each partition
        dataframe = dataframe.map_partitions(
            self.wrapped_transform,
            meta=meta_df,
        )

        # Clear divisions if component spec indicates that the index is changed
        if self._infer_index_change():
            dataframe.clear_divisions()

        return dataframe

    def _infer_index_change(self) -> bool:
        """Infer if this component changes the index based on its component spec."""
        if not self.spec.accepts_additional_subsets:
            return True
        if not self.spec.outputs_additional_subsets:
            return True
        for subset in self.spec.consumes.values():
            if not subset.additional_fields:
                return True
        return any(
            not subset.additional_fields for subset in self.spec.produces.values()
        )


class WriteComponent(Component):
    """Base class for a Fondant write component."""

    @staticmethod
    def optional_fondant_arguments() -> t.List[str]:
        return ["output_manifest_path"]

    def _load_or_create_manifest(self) -> Manifest:
        return Manifest.from_file(self.input_manifest_path)

    @abstractmethod
    def write(self, *args, **kwargs):
        """
        Abstract method to write a dataframe to a final custom location.

        Args:
            args: The dataframe will be passed in as a positional argument
            kwargs: Arguments provided to the component are passed as keyword arguments
        """

    def _process_dataset(self, manifest: Manifest) -> None:
        """
        Creates a DataLoader using the provided manifest and loads the input dataframe using the
        `load_dataframe` instance, and  applies data transformations to it using the `transform`
        method implemented by the derived class. Returns a single dataframe.

        Returns:
            A `dd.DataFrame` instance with updated data based on the applied data transformations.
        """
        data_loader = DaskDataLoader(manifest=manifest, component_spec=self.spec)
        dataframe = data_loader.load_dataframe()
        self.write(dataframe, **self.user_arguments)

    def _write_data(self, dataframe: dd.DataFrame, *, manifest: Manifest):
        """Create a data writer given a manifest and writes out the index and subsets."""

    def upload_manifest(self, manifest: Manifest, save_path: str):
        pass



"""This module defines classes to represent an Fondant component specification."""
import ast
import copy
import json
import pkgutil
import types
import typing as t
from dataclasses import dataclass
from pathlib import Path

import jsonschema.exceptions
import yaml
from jsonschema import Draft4Validator
from jsonschema.validators import RefResolver

from fondant.exceptions import InvalidComponentSpec
from fondant.schema import Field, KubeflowCommandArguments, Type

# TODO: remove after upgrading to kfpv2
kubeflow_to_python_type_dict = {
    "String": str,
    "Integer": int,
    "Float": float,
    "Boolean": ast.literal_eval,
    "JsonObject": json.loads,
    "JsonArray": json.loads,
}


def kubeflow2python_type(type_: str) -> t.Any:
    map_fn = kubeflow_to_python_type_dict[type_]
    return lambda value: map_fn(value) if value != "None" else None  # type: ignore


# TODO: Change after upgrading to kfp v2
# :https://www.kubeflow.org/docs/components/pipelines/v2/data-types/parameters/
python2kubeflow_type = {
    "str": "String",
    "int": "Integer",
    "float": "Float",
    "bool": "Boolean",
    "dict": "JsonObject",
    "list": "JsonArray",
}


@dataclass
class Argument:
    """
    Kubeflow component argument.

    Args:
        name: name of the argument
        description: argument description
        type: the python argument type (str, int, ...)
        default: default value of the argument (defaults to None)
    """

    name: str
    description: str
    type: str
    default: t.Optional[str] = None


class ComponentSubset:
    """
    Class representing a Fondant Component subset.

    Args:
        specification: the part of the component json representing the subset
    """

    def __init__(self, specification: t.Dict[str, t.Any]) -> None:
        self._specification = specification

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._specification!r})"

    @property
    def fields(self) -> t.Mapping[str, Field]:
        return types.MappingProxyType(
            {
                name: Field(name=name, type=Type.from_json(field))
                for name, field in self._specification["fields"].items()
            },
        )

    @property
    def additional_fields(self) -> bool:
        return self._specification.get("additionalFields", True)


class ComponentSpec:
    """
    Class representing a Fondant component specification.

    Args:
        specification: The fondant component specification as a Python dict
    """

    def __init__(self, specification: t.Dict[str, t.Any]) -> None:
        self._specification = copy.deepcopy(specification)
        self._validate_spec()

    def _validate_spec(self) -> None:
        """Validate a component specification against the component schema.

        Raises: InvalidComponent when the component specification is not valid.
        """
        spec_data = pkgutil.get_data("fondant", "schemas/component_spec.json")

        if spec_data is None:
            msg = "component_spec.json not found in fondant schema"
            raise FileNotFoundError(msg)

        spec_str = spec_data.decode("utf-8")
        spec_schema = json.loads(spec_str)

        base_uri = (Path(__file__).parent / "schemas").as_uri()
        resolver = RefResolver(base_uri=f"{base_uri}/", referrer=spec_schema)
        validator = Draft4Validator(spec_schema, resolver=resolver)

        try:
            validator.validate(self._specification)
        except jsonschema.exceptions.ValidationError as e:
            raise InvalidComponentSpec.create_from(e)

    @classmethod
    def from_file(cls, path: t.Union[str, Path]) -> "ComponentSpec":
        """Load the component spec from the file specified by the provided path."""
        with open(path, encoding="utf-8") as file_:
            specification = yaml.safe_load(file_)
            return cls(specification)

    def to_file(self, path) -> None:
        """Dump the component spec to the file specified by the provided path."""
        with open(path, "w", encoding="utf-8") as file_:
            yaml.dump(self._specification, file_)

    @property
    def name(self):
        return self._specification["name"]

    @property
    def description(self):
        return self._specification["description"]

    @property
    def image(self):
        return self._specification["image"]

    @property
    def index(self):
        return ComponentSubset({"fields": {}})

    @property
    def consumes(self) -> t.Mapping[str, ComponentSubset]:
        """The subsets consumed by the component as an immutable mapping."""
        return types.MappingProxyType(
            {
                name: ComponentSubset(subset)
                for name, subset in self._specification.get("consumes", {}).items()
                if name != "additionalSubsets"
            },
        )

    @property
    def produces(self) -> t.Mapping[str, ComponentSubset]:
        """The subsets produced by the component as an immutable mapping."""
        return types.MappingProxyType(
            {
                name: ComponentSubset(subset)
                for name, subset in self._specification.get("produces", {}).items()
                if name != "additionalSubsets"
            },
        )

    @property
    def accepts_additional_subsets(self) -> bool:
        return self._specification.get("consumes", {}).get("additionalSubsets", True)

    @property
    def outputs_additional_subsets(self) -> bool:
        return self._specification.get("produces", {}).get("additionalSubsets", True)

    @property
    def args(self) -> t.Dict[str, Argument]:
        return {
            name: Argument(
                name=name,
                description=arg_info["description"],
                type=arg_info["type"],
                default=arg_info["default"] if "default" in arg_info else None,
            )
            for name, arg_info in self._specification.get("args", {}).items()
        }

    @property
    def specification(self) -> t.Dict[str, t.Any]:
        return copy.deepcopy(self._specification)

    @property
    def kubeflow_specification(self) -> "KubeflowComponentSpec":
        return KubeflowComponentSpec.from_fondant_component_spec(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._specification!r})"

    def __eq__(self, other):
        if not isinstance(other, ComponentSpec):
            return False
        return self._specification == other._specification


class KubeflowComponentSpec:
    """
    Class representing a Kubeflow component specification.

    Args:
        specification: The kubeflow component specification as a Python dict
    """

    def __init__(self, specification: t.Dict[str, t.Any]) -> None:
        self._specification = specification

    @classmethod
    def from_fondant_component_spec(
        cls,
        fondant_component: ComponentSpec,
    ) -> "KubeflowComponentSpec":
        """Create a Kubeflow component spec from a Fondant component spec."""
        specification = {
            "name": fondant_component.name,
            "description": fondant_component.description,
            "inputs": [
                {
                    "name": "input_manifest_path",
                    "description": "Path to the input manifest",
                    "type": "String",
                },
                {
                    "name": "metadata",
                    "description": "Metadata arguments containing the run id and base path",
                    "type": "String",
                },
                {
                    "name": "component_spec",
                    "description": "The component specification as a dictionary",
                    "type": "JsonObject",
                    "default": "None",
                },
                *(
                    {
                        "name": arg.name,
                        "description": arg.description,
                        "type": python2kubeflow_type[arg.type],
                        **({"default": arg.default} if arg.default is not None else {}),
                    }
                    for arg in fondant_component.args.values()
                ),
            ],
            "outputs": [
                {
                    "name": "output_manifest_path",
                    "description": "Path to the output manifest",
                    "type": "String",
                },
            ],
            "implementation": {
                "container": {
                    "image": fondant_component.image,
                    "command": [
                        "python3",
                        "main.py",
                        "--input_manifest_path",
                        {"inputPath": "input_manifest_path"},
                        "--metadata",
                        {"inputValue": "metadata"},
                        "--component_spec",
                        {"inputValue": "component_spec"},
                        *cls._dump_args(fondant_component.args.values()),
                        "--output_manifest_path",
                        {"outputPath": "output_manifest_path"},
                    ],
                },
            },
        }
        return cls(specification)

    @staticmethod
    def _dump_args(args: t.Iterable[Argument]) -> KubeflowCommandArguments:
        """Dump Fondant specification arguments to kfp command arguments."""
        dumped_args: KubeflowCommandArguments = []
        for arg in args:
            arg_name = arg.name.strip().replace(" ", "_")
            arg_name_cmd = f"--{arg_name}"

            dumped_args.append(arg_name_cmd)
            dumped_args.append({"inputValue": arg_name})

        return dumped_args

    def to_file(self, path: t.Union[str, Path]) -> None:
        """Dump the component specification to the file specified by the provided path."""
        with open(path, "w", encoding="utf-8") as file_:
            yaml.dump(
                self._specification,
                file_,
                indent=4,
                default_flow_style=False,
                sort_keys=False,
            )

    def to_string(self) -> str:
        """Return the component specification as a string."""
        return json.dumps(self._specification)

    @property
    def input_arguments(self) -> t.Mapping[str, Argument]:
        """The input arguments of the component as an immutable mapping."""
        return types.MappingProxyType(
            {
                info["name"]: Argument(
                    name=info["name"],
                    description=info["description"],
                    type=info["type"],
                    default=info["default"] if "default" in info else None,
                )
                for info in self._specification["inputs"]
            },
        )

    @property
    def output_arguments(self) -> t.Mapping[str, Argument]:
        """The output arguments of the component as an immutable mapping."""
        return types.MappingProxyType(
            {
                info["name"]: Argument(
                    name=info["name"],
                    description=info["description"],
                    type=info["type"],
                )
                for info in self._specification["outputs"]
            },
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._specification!r})"



"""This component filters code based on a set of metadata associated with it."""
import logging

import dask.dataframe as dd

from fondant.component import DaskTransformComponent

logger = logging.getLogger(__name__)


class FilterLineLengthComponent(DaskTransformComponent):
    """
    This component filters code based on a set of metadata associated with it:
    average line length, maximum line length and alphanum fraction.
    """

    def transform(
        self,
        *,
        dataframe: dd.DataFrame,
        avg_line_length_threshold: int,
        max_line_length_threshold: int,
        alphanum_fraction_threshold: float,
    ) -> dd.DataFrame:
        """
        Args:
            dataframe: Dask dataframe
            avg_line_length_threshold: Threshold for average line length to filter on
            max_line_length_threshold: Threshold for max line length to filter on
            alphanum_fraction_threshold: Alphanum fraction to filter on
        Returns:
            Filtered dask dataframe.
        """
        return dataframe[
            (dataframe["code_avg_line_length"] > avg_line_length_threshold)
            & (dataframe["code_max_line_length"] > max_line_length_threshold)
            & (dataframe["code_alphanum_fraction"] > alphanum_fraction_threshold)
        ]


if __name__ == "__main__":
    component = FilterLineLengthComponent.from_args()
    component.run()



"""This component filters images of the dataset based on image size (minimum height and width)."""
import io
import logging
import typing as t

import imagesize
import numpy as np
import pandas as pd

from fondant.component import PandasTransformComponent

logger = logging.getLogger(__name__)


def extract_dimensions(images: bytes) -> t.Tuple[np.int16, np.int16]:
    """Extract the width and height of an image.

    Args:
        images: input dataframe with images_data column

    Returns:
        np.int16: width of the image
        np.int16: height of the image
    """
    width, height = imagesize.get(io.BytesIO(images))

    return np.int16(width), np.int16(height)


class ImageResolutionExtractionComponent(PandasTransformComponent):
    """Component that extracts image dimensions."""

    def transform(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """
        Args:
            dataframe: Dask dataframe
        Returns:
            dataset.
        """
        logger.info("Filtering dataset...")

        dataframe[[("images", "width"), ("images", "height")]] = dataframe[
            [("images", "data")]
        ].apply(lambda x: extract_dimensions(x.images.data), axis=1)

        return dataframe


if __name__ == "__main__":
    component = ImageResolutionExtractionComponent.from_args()
    component.run()



"""
This component downloads images based on URLs, and resizes them based on various settings like
minimum image size and aspect ratio.

Some functions here are directly taken from
https://github.com/rom1504/img2dataset/blob/main/img2dataset/downloader.py.
"""
import io
import logging
import traceback
import urllib

import pandas as pd
from resizer import Resizer

from fondant.component import PandasTransformComponent

logger = logging.getLogger(__name__)


def is_disallowed(headers, user_agent_token, disallowed_header_directives):
    """Check if HTTP headers contain an X-Robots-Tag directive disallowing usage."""
    for values in headers.get_all("X-Robots-Tag", []):
        try:
            uatoken_directives = values.split(":", 1)
            directives = [x.strip().lower() for x in uatoken_directives[-1].split(",")]
            ua_token = (
                uatoken_directives[0].lower() if len(uatoken_directives) == 2  # noqa: PLR2004
                else None
            )
            if (ua_token is None or ua_token == user_agent_token) and any(
                x in disallowed_header_directives for x in directives
            ):
                return True
        except Exception as err:
            traceback.print_exc()
            print(f"Failed to parse X-Robots-Tag: {values}: {err}")
    return False


def download_image(url, timeout, user_agent_token, disallowed_header_directives):
    """Download an image with urllib."""
    img_stream = None
    user_agent_string = (
        "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:72.0) Gecko/20100101 Firefox/72.0"
    )
    if user_agent_token:
        user_agent_string += f" (compatible; {user_agent_token}; " \
                             f"+https://github.com/rom1504/img2dataset)"
    try:
        request = urllib.request.Request(
            url, data=None, headers={"User-Agent": user_agent_string},
        )
        with urllib.request.urlopen(request, timeout=timeout) as r:
            if disallowed_header_directives and is_disallowed(
                r.headers,
                user_agent_token,
                disallowed_header_directives,
            ):
                return None
            img_stream = io.BytesIO(r.read())
        return img_stream
    except Exception:
        if img_stream is not None:
            img_stream.close()
        return None


def download_image_with_retry(
    url,
    *,
    timeout,
    retries,
    resizer,
    user_agent_token=None,
    disallowed_header_directives=None,
):
    for _ in range(retries + 1):
        img_stream = download_image(
            url, timeout, user_agent_token, disallowed_header_directives,
        )
        if img_stream is not None:
            # resize the image
            return resizer(img_stream)
    return None, None, None


class DownloadImagesComponent(PandasTransformComponent):
    """Component that downloads images based on URLs."""

    def setup(
        self,
        *,
        timeout: int = 10,
        retries: int = 0,
        image_size: int = 256,
        resize_mode: str = "border",
        resize_only_if_bigger: bool = False,
        min_image_size: int = 0,
        max_aspect_ratio: float = float("inf"),
    ):
        logger.info("Instantiating resizer...")
        self.resizer = Resizer(
            image_size=image_size,
            resize_mode=resize_mode,
            resize_only_if_bigger=resize_only_if_bigger,
            min_image_size=min_image_size,
            max_aspect_ratio=max_aspect_ratio,
        )
        self.timeout = timeout
        self.retries = retries

    def transform(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        dataframe[[
            ("images", "data"),
            ("images", "width"),
            ("images", "height"),
        ]] = dataframe.apply(
            lambda example: download_image_with_retry(
                url=example["images"]["url"],
                timeout=self.timeout,
                retries=self.retries,
                resizer=self.resizer,
            ),
            axis=1,
            result_type="expand",
        )

        return dataframe


if __name__ == "__main__":
    component = DownloadImagesComponent.from_args()
    component.run()



import json

from utils.emails_ip_addresses_detection import detect_email_addresses
from utils.keys_detection import detect_keys


def postprocess_secrets(secrets):
    """Postprocess the secrets found by the scan_secrets function."""
    if secrets:
        matches = json.dumps(secrets)
        has_secrets = True
    else:
        matches = json.dumps([])
        has_secrets = False
    return matches, has_secrets


def scan_pii(text, key_detector="other"):
    """Scan a piece of code to detect PII
    This add 3 columns to the dataset:
    - secrets: (list) of secrets/PII found
    - has_secrets: (bool) whether the example contains secrets/PII
    - num_secrests (int) number of secrets.
    """
    secrets = []
    if key_detector == "regex":
        # use a regex to detect keys + emails + ips
        secrets = secrets + detect_email_addresses(
            text, tag_types={"KEY", "EMAIL", "IP_ADDRESS"},
        )
    else:
        # detect emails and ip addresses with regexes
        secrets = secrets + detect_email_addresses(
            text, tag_types={"EMAIL", "IP_ADDRESS"},
        )
        # for keys use detect-secrets tool
        secrets = secrets + detect_keys(text)
    # to add this as new columns to datasets we need the same number of samples in each row
    # we save secrets as json strings instead of lists
    matches, has_secrets = postprocess_secrets(secrets)

    return matches, has_secrets, len(secrets)



"""This component that embeds images using a model from the Hugging Face hub."""
import io
import logging
import typing as t

import numpy as np
import pandas as pd
import torch
from PIL import Image
from transformers import CLIPProcessor, CLIPVisionModelWithProjection

from fondant.component import PandasTransformComponent

logger = logging.getLogger(__name__)


def process_image(image: bytes, *, processor: CLIPProcessor, device: str) -> torch.Tensor:
    """
    Process the image to a tensor.

    Args:
        image: The input image as a byte string.
        processor: The processor object for transforming the image.
        device: The device to move the transformed image to.
    """
    def load(img: bytes) -> Image:
        """Load the bytestring as an image."""
        bytes_ = io.BytesIO(img)
        return Image.open(bytes_).convert("RGB")

    def transform(img: Image) -> torch.Tensor:
        """Transform the image to a tensor using a clip processor and move it to the specified
        device.
        """
        return processor(images=img, return_tensors="pt").to(device)

    return transform(load(image))["pixel_values"]


@torch.no_grad()
def embed_image_batch(image_batch: pd.DataFrame, *, model: CLIPVisionModelWithProjection) -> \
        pd.Series:
    """Embed a batch of images."""
    input_batch = torch.cat(image_batch.tolist())
    output_batch = model(input_batch)
    embeddings_batch = output_batch.image_embeds.cpu().tolist()
    return pd.Series(embeddings_batch, index=image_batch.index)


class EmbedImagesComponent(PandasTransformComponent):
    """Component that captions images using a model from the Hugging Face hub."""

    def setup(
        self,
        *,
        model_id: str,
        batch_size: int,
    ):
        """
        Args:
            model_id: id of the model on the Hugging Face hub
            batch_size: batch size to use.
        """
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        logger.info("device used is %s", self.device)

        logger.info("Initialize model '%s'", model_id)
        self.processor = CLIPProcessor.from_pretrained(model_id)
        self.model = CLIPVisionModelWithProjection.from_pretrained(model_id).to(self.device)
        logger.info("Model initialized")

        self.batch_size = batch_size

    def transform(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        images = dataframe["images"]["data"].apply(
            process_image,
            processor=self.processor,
            device=self.device,
        )
        results: t.List[pd.Series] = []
        for batch in np.split(images, np.arange(self.batch_size, len(images), self.batch_size)):
            if not batch.empty:
                results.append(embed_image_batch(batch, model=self.model).T)

        return pd.concat(results).to_frame(name=("embeddings", "data"))


if __name__ == "__main__":
    component = EmbedImagesComponent.from_args()
    component.run()



"""This component writes an image dataset to the hub."""
import logging
import typing as t
from io import BytesIO

import dask.dataframe as dd
import datasets

# Define the schema for the struct using PyArrow
import huggingface_hub
from PIL import Image

from fondant.component import WriteComponent

logger = logging.getLogger(__name__)


def convert_bytes_to_image(image_bytes: bytes, feature_encoder: datasets.Image) -> \
        t.Dict[str, t.Any]:
    """
    Function that converts image bytes to hf image format
    Args:
        image_bytes: the images as a bytestring
        feature_encoder: hf image feature encoder
    Returns:
        HF image representation.
    """
    image = Image.open(BytesIO(image_bytes))
    image = feature_encoder.encode_example(image)
    return image


class WriteToHubComponent(WriteComponent):
    def write(
            self,
            dataframe: dd.DataFrame,
            *,
            hf_token: str,
            username: str,
            dataset_name: str,
            image_column_names: t.Optional[list],
            column_name_mapping: t.Optional[dict],
    ):
        """
        Args:
            dataframe: Dask dataframe
            hf_token: The hugging face token used to write to the hub
            username: The username under which to upload the dataset
            dataset_name: The name of the dataset to upload
            image_column_names: A list containing the subset image column names. Used to format the
            image fields to HF hub format
            column_name_mapping: Mapping of the consumed fondant column names to the written hub
             column names.
        """
        # login
        huggingface_hub.login(token=hf_token)

        # Create HF dataset repository
        repo_id = f"{username}/{dataset_name}"
        repo_path = f"hf://datasets/{repo_id}"
        logger.info(f"Creating HF dataset repository under ID: '{repo_id}'")
        huggingface_hub.create_repo(repo_id=repo_id, repo_type="dataset", exist_ok=True)

        # Get columns to write and schema
        write_columns = []
        schema_dict = {}
        for subset_name, subset in self.spec.consumes.items():
            for field in subset.fields.values():
                column_name = f"{subset_name}_{field.name}"
                write_columns.append(column_name)
                if image_column_names and column_name in image_column_names:
                    schema_dict[column_name] = datasets.Image()
                else:
                    schema_dict[column_name] = datasets.Value(str(field.type.value))

        schema = datasets.Features(schema_dict).arrow_schema
        dataframe = dataframe[write_columns]

        # Map image column to hf data format
        feature_encoder = datasets.Image(decode=True)

        if image_column_names is not None:
            for image_column_name in image_column_names:
                dataframe[image_column_name] = dataframe[image_column_name].map(
                    lambda x: convert_bytes_to_image(x, feature_encoder),
                    meta=(image_column_name, "object"),
                )

        # Map column names to hf data format
        if column_name_mapping:
            dataframe = dataframe.rename(columns=column_name_mapping)

        # Write dataset to the hub
        dd.to_parquet(dataframe, path=f"{repo_path}/data", schema=schema)


if __name__ == "__main__":
    component = WriteToHubComponent.from_args()
    component.run()



"""A component that detects and redacts Personal Identifiable Information (PII) in code."""

import json
import logging

import dask.dataframe as dd
from pii_detection import scan_pii
from pii_redaction import redact_pii

from fondant.component import DaskTransformComponent

logger = logging.getLogger(__name__)


class RemovePIIComponent(DaskTransformComponent):
    """Component that detects and redacts PII from code."""

    def transform(
        self,
        *,
        dataframe: dd.DataFrame,
    ) -> dd.DataFrame:
        """
        Args:
            dataframe: Dask dataframe.

        Returns:
            Dask dataframe
        """
        # detect PII
        result = dataframe.apply(
            lambda example: scan_pii(text=example.code_content),
            axis=1,
            result_type="expand",
            meta={0: object, 1: bool, 2: int},
        )
        result.columns = ["code_secrets", "code_has_secrets", "code_number_secrets"]

        dataframe = dataframe.merge(result, left_index=True, right_index=True)

        # redact PII
        # we use random replacements by default
        with open("replacements.json") as f:
            replacements = json.load(f)

        dataframe["code_content"] = dataframe.apply(
            lambda example: redact_pii(
                text=example.code_content,
                secrets=example.code_secrets,
                has_secrets=example.code_has_secrets,
                replacements=replacements,
            ),
            axis=1,
            meta=(None, "str"),
        )
        dataframe = dataframe.drop(
            ["code_secrets", "code_has_secrets", "code_number_secrets"], axis=1,
        )

        return dataframe


if __name__ == "__main__":
    component = RemovePIIComponent.from_args()
    component.run()



"""This code is adapted from BigScience PII detection
https://github.com/bigcode-project/bigcode-dataset/blob/main/preprocessing/filtering.py.

MST BigScience PII Code
Original colab that is a source of this file is located at
    https://colab.research.google.com/drive/1086H3-LGMz3gX0pGy9ECgr8KflosSKso
# License
Copyright 2022 Authors of this Notebook
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import ast
import io
import tokenize
import typing as t
import warnings
from itertools import groupby

NODE_TYPES = {
    ast.ClassDef: "Class",
    ast.FunctionDef: "Function/Method",
    ast.Module: "Module",
}


# Note: sometimes this can miss examples with decorators over classes
# ast parsing, source: https://gist.github.com/SpotlightKid/1548cb6c97f2a844f72d
def parse_docstrings(source):
    """
    Parse Python source code and yield a tuple of ast node instance, name,
    and docstring for each function/method, class and module.

    Args:
        source: The Python source code to parse.

    Yields:
        A tuple containing ast node instance, name, and docstring.
    """
    tree = ast.parse(source)

    for node in ast.walk(tree):
        if isinstance(node, tuple(NODE_TYPES)):
            docstring = ast.get_docstring(node)

            yield node, getattr(node, "name", None), docstring


# comment extraction
def get_comments(source: str) -> str:
    """
    Returns a string including all comments in python code.

    Args:
        source: the code to parse

    Returns:
        The script comments.
    """
    comments = []
    g = tokenize.generate_tokens(io.StringIO(source).readline)
    for toknum, tokval, _, _, _ in g:
        if toknum == tokenize.COMMENT:
            comments.append((toknum, tokval))
    return tokenize.untokenize(comments).replace("#", "")


def get_docstrings(source: str) -> t.List[str]:
    """
    Parse Python source code from file or string and print docstrings.

    Args:
        source: the code to parse

    Returns:
        A list containing the script docstrings.
    """
    if hasattr(source, "read"):
        source = source.read()

    docstrings = sorted(
        parse_docstrings(source), key=lambda x: (NODE_TYPES.get(type(x[0])), x[1]),
    )

    grouped = groupby(docstrings, key=lambda x: NODE_TYPES.get(type(x[0])))
    results = []
    for _, group in grouped:
        for _, _name, docstring in group:
            if docstring:
                results.append(docstring)
    return results


def get_text_python(source: str, extract_comments: bool = True) -> str:
    """Extract all natural text in source: comments + docstrings
    the extraction fails in case of syntax errors in the file.

    Args:
        source: the code to parse
        extract_comments: if True extract comments too

    Returns:
        A string with concatenated docstrings and comments.
    """
    try:
        docstrings = "\n".join(get_docstrings(source))
    except Exception:
        docstrings = ""
        warnings.warn(
            "code couldn't be parsed due to compilation failure, no docstring is extracted",
        )

    if extract_comments:
        try:
            comments = get_comments(source)
        except Exception:
            comments = ""
            warnings.warn("tokenization error, no comments are extracted")
    else:
        comments = ""

    output = docstrings + "\n" + comments
    return output.strip()


def get_comments_to_code_ratio(text: str) -> float:
    """
    Get the ratio of comments to code in a program
    Args:
        text: the string source code
    Returns:
        The comments to code ratio.
    """
    comments = get_text_python(text)

    return len(comments) / len(text)



"""This code is adapted from BigScience PII detection
https://github.com/bigscience-workshop/data-preparation/blob/main/preprocessing/training/02_pii/bigscience_pii_detect_redact.py.

MST BigScience PII Code
Original colab that is a source of this file is located at
    https://colab.research.google.com/drive/1086H3-LGMz3gX0pGy9ECgr8KflosSKso
# License
Copyright 2022 Authors of this Notebook
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import ipaddress
import sys

import regex
from gibberish_detector import detector

# Regexes for PII detection
# ruff: noqa: E501

year_patterns = [
    regex.compile(
        r"(?:^|[\b\s@?,!;:\'\")(.\p{Han}])([1-2][0-9]{3}[\p{Pd}/][1-2][0-9]{3})(?:$|[\s@,?!;:\'\"(.\p{Han}])",
    ),  # yyyy-yyyy or yyyy/yyyy
    regex.compile(
        r"(?:^|[\b\s@?,!;:\'\")(.\p{Han}])([1-2][0-9]{3}[\p{Pd}/.][0-3][0-9][\p{Pd}/.][0-3][0-9])(?:$|[\s@,?!;:\'\"(.\p{Han}])",
    ),  # yyyy-mm-dd or yyyy-dd-mm or yyyy/mm/dd or yyyy/dd/mm or yyyy.mm.dd or yyyy.dd.mm
    regex.compile(
        r"(?:^|[\b\s@?,!;:\'\")(.\p{Han}])([0-3][0-9][\p{Pd}/.][0-3][0-9][\p{Pd}/.](?:[0-9]{2}|[1-2][0-9]{3}))(?:$|[\s@,?!;:\'\"(.\p{Han}])",
    ),  # mm-dd-yyyy or dd-mm-yyyy or mm/dd/yyyy or dd/mm/yyyy or mm.dd.yyyy or dd.mm.yyyy or the same but with yy instead of yyyy
    regex.compile(
        r"(?:^|[\b\s@?,!;:\'\")(.\p{Han}])([0-3][0-9][\p{Pd}/](?:[0-9]{2}|[1-2][0-9]{3}))(?:$|[\s@,?!;:\'\"(.\p{Han}])",
    ),  # mm-yyyy or mm/yyyy or the same but with yy
    regex.compile(
        r"(?:^|[\b\s@?,!;:\'\")(.\p{Han}])([1-2][0-9]{3}-[0-3][0-9])(?:$|[\s@,?!;:\'\"(.\p{Han}])",
    ),  # yyyy-mm or yyyy/mm
]

key_pattern = r"((?:(?:[A-Za-z]+[\p{Nd}\p{Pd}\/\+\=:_]+|[\p{Nd}\p{Pd}\/\+\=:]+[A-Za-z]+)){10,})(?:$|[\b\s\p{Han}@?,!;:\'\")(.])"
key_pattern_2 = r"(?!(?:\/[^ .]+){2,})((?:(?:[A-Za-z]+[\p{Nd}\p{Pd}\/\+\=:_]+|[\p{Nd}\p{Pd}\/\+\=:]+[A-Za-z]+)){4,})(?:$|[\b\s\p{Han}@?,!;:\'\")(.])"
ipv4_pattern = r"(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)(?:\.(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)){3}"
ipv6_pattern = r"(?:[0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|(?:[0-9a-fA-F]{1,4}:){1,7}:|(?:[0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|(?:[0-9a-fA-F]{1,4}:){1,5}(?::[0-9a-fA-F]{1,4}){1,2}|(?:[0-9a-fA-F]{1,4}:){1,4}(?::[0-9a-fA-F]{1,4}){1,3}|(?:[0-9a-fA-F]{1,4}:){1,3}(?::[0-9a-fA-F]{1,4}){1,4}|(?:[0-9a-fA-F]{1,4}:){1,2}(?::[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:(?:(?::[0-9a-fA-F]{1,4}){1,6})|:(?:(?::[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(?::[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(?:ffff(?::0{1,4}){0,1}:){0,1}(?:(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])|(?:[0-9a-fA-F]{1,4}:){1,4}:(?:(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])"
ip_pattern = (
    r"(?:^|[\b\s@?,!;:\'\")(.\p{Han}])("
    + r"|".join([ipv4_pattern, ipv6_pattern])
    + ")(?:$|[\\s@,?!;:'\"(.\\p{Han}])"
)

# Note: to reduce false positives, a number of technically-valid-but-rarely-used
# email address patterns (e.g. with parenthesis or slashes) will not match
email_pattern = r"""
    (?<= ^ | [[({<\b\s@,?!;'"\p{Han}¿¡:.] | \\['"] )  # left delimiter
    (
      (?:                                             # local part
        [^][(){}<>\b\s@,?!;'":#/\\=.\-]               # arbitrary character
        |
        (?: [=.\-] (?! [.@]) )                        # ".=-" not before ".@"
      )+
      @
      (?:
        (?:
             \w                                       # single-letter subdomain
           |
             [^.\b\s@?!;,/()>\-:]                     # subdomain (>=2 letter)
             [^.\b\s@?!;,/()>]{0,62}
             [^.\b\s@?!;,/()>\-:'"]
        )
        \.
      ){1,10}
      (?: [\p{L}\p{M}]{2,63} | xn-- \w+ )             # TLD, including IDN
    )
    (?= $ | [])}>\b\s@,?!;'"\p{Han}] | \\['"] | : (?! \d) | \. (?! \S))   # right delim
"""


def get_regexes(high_risk_tags={"EMAIL", "IP_ADDRESS", "KEY"}):
    """Returns a dict of regexes to match the PII in high_risk_tags."""
    key_regex = regex.compile(key_pattern_2, flags=regex.MULTILINE)
    ipv4_regex = regex.compile(ipv4_pattern)
    ipv6_regex = regex.compile(ipv6_pattern)
    ip_regex = regex.compile(ip_pattern, flags=regex.MULTILINE)
    email_regex = regex.compile(email_pattern, flags=regex.MULTILINE | regex.VERBOSE)

    mst_regexes = {}
    for tag in high_risk_tags:
        if tag == "KEY":
            mst_regexes["KEY"] = key_regex
        elif tag == "IPv4":
            mst_regexes["IPv4"] = ipv4_regex
        elif tag == "IPv6":
            mst_regexes["IPv6"] = ipv6_regex
        elif tag == "IP_ADDRESS":
            mst_regexes["IP_ADDRESS"] = ip_regex
        elif tag == "EMAIL":
            mst_regexes["EMAIL"] = email_regex
        else:
            sys.stderr.write("Don't have tag regex pattern for %s =(" % tag)

    return mst_regexes


def ip_has_digit(matched_str):
    """Checks to make sure the PII span is not just :: or whatever that may
    accidentally be picked up by making sure there are digits.
    """
    return any(map(str.isdigit, matched_str))


def matches_date_pattern(matched_str):
    # Screen out date false positives
    return any(year_regex.match(matched_str) for year_regex in year_patterns)


def filter_versions(matched_str, context):
    """Filter addresses in this format x.x.x.x  and the words dns/server
    don't appear in the neighboring context, usually they are just versions.
    """
    # count occurrence of dots
    dot_count = matched_str.count(".")
    exclude = dot_count == 3 and len(matched_str) == 7  # noqa: PLR2004 (magic value)
    if exclude and ("dns" in context.lower() or "server" in context.lower()):
        return False
    return exclude


def not_ip_address(matched_str):
    """Make sure the string has a valid IP address format
    e.g: 33.01.33.33 is not a valid IP address because of the 0 in front of 1
    TODO: fix this directly in the regex.
    """
    try:
        ipaddress.ip_address(matched_str)
        return False
    except ValueError:
        return True


def is_gibberish(matched_str):
    """Checks to make sure the PII span is gibberish and not word like."""
    # pip install gibberish-detector
    # download the training corpora from
    # https://raw.githubusercontent.com/domanchi/gibberish-detector/master/examples/big.txt
    # run gibberish-detector train big.txt > big.model to generate the model (it takes 3 seconds)
    Detector = detector.create_from_model("gibberish_data/big.model")
    return Detector.is_gibberish(matched_str.lower())


def detect_email_addresses(content, tag_types={"EMAIL", "IP_ADDRESS"}):
    """Detects email addresses in a string using regex matching
    Args:
      content (str): A string containing the text to be analyzed.
      tag_types (set): A set of tag types to be detected. Defaults to EMAIL and IP_ADDRESS.
      you can add 'KEY' to detect keys with a regex.

    Returns:
        A list of dicts containing the tag type, the matched string, and the start and
        end indices of the match.
    """
    mst_regexes = get_regexes(tag_types)
    matches = []
    for tag in tag_types:
        label_pattern = mst_regexes[tag]
        # regex matching
        matches_tmp = label_pattern.finditer(content)
        for match in matches_tmp:
            if match.groups():
                if len(match.groups()) > 1 and match.groups()[1]:
                    sys.stderr.write(
                        "Warning: Found substring matches in the main match.",
                    )
                # setup outputs
                value = match.group(1)
                start, end = match.span(1)
                if value:
                    if tag == "IP_ADDRESS":
                        # Filter out false positive IPs
                        if not ip_has_digit(value):
                            continue
                        if matches_date_pattern(value):
                            continue
                        if filter_versions(
                            value, content[start - 100 : end + 100],
                        ) or not_ip_address(value):
                            continue
                        # combine if conditions in one

                    # Filter out false positive keys
                    if tag == "KEY" and not is_gibberish(value):
                        continue
                    matches.append(
                        {
                            "tag": tag,
                            "value": value,
                            "start": start,
                            "end": end,
                        },
                    )
                else:
                    msg = "No match found inside groups"
                    raise ValueError(msg)
    return matches



"""This component loads a seed dataset from the hub."""
import logging
import typing as t

import dask.dataframe as dd

from fondant.component import LoadComponent

logger = logging.getLogger(__name__)


class LoadFromHubComponent(LoadComponent):
    def load(self,
             *,
             dataset_name: str,
             column_name_mapping: dict,
             image_column_names: t.Optional[list],
             n_rows_to_load: t.Optional[int]) -> dd.DataFrame:
        """
        Args:
            dataset_name: name of the dataset to load.
            column_name_mapping: Mapping of the consumed hub dataset to fondant column names
            image_column_names: A list containing the original hub image column names. Used to
                format the image from HF hub format to a byte string
            n_rows_to_load: optional argument that defines the number of rows to load. Useful for
              testing pipeline runs on a small scale
        Returns:
            Dataset: HF dataset.
        """
        # 1) Load data, read as Dask dataframe
        logger.info("Loading dataset from the hub...")
        dask_df = dd.read_parquet(f"hf://datasets/{dataset_name}")

        # 2) Make sure images are bytes instead of dicts
        if image_column_names is not None:
            for image_column_name in image_column_names:
                dask_df[image_column_name] = dask_df[image_column_name].map(
                    lambda x: x["bytes"], meta=("bytes", bytes),
                )

        # 3) Rename columns
        dask_df = dask_df.rename(columns=column_name_mapping)

        # 4) Optional: only return specific amount of rows

        if n_rows_to_load:
            dask_df = dask_df.head(n_rows_to_load)
            dask_df = dd.from_pandas(dask_df, npartitions=1)

        return dask_df


if __name__ == "__main__":
    component = LoadFromHubComponent.from_args()
    component.run()



"""This component crops images by removing the borders."""
import io
import logging
import typing as t

import dask.dataframe as dd
import numpy as np
from image_crop import remove_borders
from PIL import Image

from fondant.component import DaskTransformComponent

logger = logging.getLogger(__name__)


def extract_dimensions(image_df: dd.DataFrame) -> t.Tuple[np.int16, np.int16]:
    """Extract the width and height of an image.

    Args:
        image_df (dd.DataFrame): input dataframe with images_data column

    Returns:
        np.int16: width of the image
        np.int16: height of the image
    """
    image = Image.open(io.BytesIO(image_df["images_data"]))

    return np.int16(image.size[0]), np.int16(image.size[1])


class ImageCroppingComponent(DaskTransformComponent):
    """Component that crops images."""

    def transform(
        self,
        *,
        dataframe: dd.DataFrame,
        cropping_threshold: int = -30,
        padding: int = 10,
    ) -> dd.DataFrame:
        """
        Args:
            dataframe (dd.DataFrame): Dask dataframe
            cropping_threshold (int): threshold parameter used for detecting borders
            padding (int): padding for the image cropping.

        Returns:
            dd.DataFrame: Dask dataframe with cropped images
        """
        # crop images
        dataframe["images_data"] = dataframe["images_data"].map(
            lambda x: remove_borders(x, cropping_threshold, padding),
            meta=("images_data", "bytes"),
        )

        # extract width and height
        dataframe[["images_width", "images_height"]] = dataframe[
            [
                "images_data",
            ]
        ].apply(extract_dimensions, axis=1, result_type="expand", meta={0: int, 1: int})

        return dataframe


if __name__ == "__main__":
    component = ImageCroppingComponent.from_args()
    component.run()



"""This file contains the image cropping logic."""
import io
import typing as t

import numpy as np
from PIL import Image, ImageChops, ImageOps


def most_common(lst: t.Sequence[t.Tuple]) -> t.Tuple:
    """Get the most common element from a list.

    Args:
        lst (List): input list

    Returns:
        object: most common element
    """
    return max(set(lst), key=lst.count)


def get_image_borders(image: Image.Image) -> t.Tuple:
    """Get the most common color from the image borders.

    Args:
        image (Image.Image): input image

    Returns:
        tuple: most common color
    """
    image = np.array(image)

    # get the colors of the image border
    border_colors = [
        *image[:, 0, :].tolist(),
        *image[:, -1, :].tolist(),
        *image[0, :, :].tolist(),
        *image[-1, :, :].tolist(),
    ]

    # map the colors to RGB tuples
    rgb_border_colors = tuple(map(tuple, border_colors))

    # calculate most common RGB color
    color_common = most_common(rgb_border_colors)

    return tuple(color_common)


def remove_borders(
    image_bytes: bytes, cropping_threshold: int = -30, padding: int = 10,
) -> bytes:
    """This method removes borders by checking the overlap between
    a color and the original image. By subtracting these two
    it finds the minimal bounding box.

    Args:
        image_bytes (bytes): input images in bytes
        cropping_threshold (int): threshold parameter used for detecting borders
        padding (int): padding for the image cropping

    Returns:
      bytes: cropped image
    """
    image = Image.open(io.BytesIO(image_bytes)).convert("RGB")
    width, height = image.size

    color_common = get_image_borders(image)
    background = Image.new(image.mode, image.size, color_common)

    diff = ImageChops.difference(image, background)
    diff = ImageChops.add(diff, diff, 1.0, cropping_threshold)

    bbox = diff.getbbox()
    if not bbox:
        return image_bytes

    bbox = (
        np.max([0, bbox[0] - padding]),
        np.max([0, bbox[1] - padding]),
        np.min([width, bbox[2] + padding]),
        np.min([height, bbox[3] + padding]),
    )

    image_crop = image.crop(bbox)

    # if the image is not square, add padding to the shorter side
    # in color `color_common`
    if image_crop.size[0] != image_crop.size[1]:
        if image_crop.size[0] > image_crop.size[1]:
            padding = int((image_crop.size[0] - image_crop.size[1]) / 2)
            image_crop = ImageOps.expand(
                image_crop, border=(0, padding), fill=color_common,
            )
        else:
            padding = int((image_crop.size[1] - image_crop.size[0]) / 2)
            image_crop = ImageOps.expand(
                image_crop, border=(padding, 0), fill=color_common,
            )

    # serialize image to JPEG
    crop_bytes = io.BytesIO()
    image_crop.save(crop_bytes)

    return crop_bytes.getvalue()



import subprocess  # nosec
from abc import ABC, abstractmethod


class Runner(ABC):
    """Abstract base class for a runner."""

    @abstractmethod
    def run(self, *args, **kwargs):
        """Abstract method to invoke running."""


class DockerRunner(Runner):
    def run(cls, input_spec: str, *args, **kwargs):
        """Run a docker-compose spec."""
        cmd = ["docker", "compose", "-f", input_spec, "up", "--build"]

        subprocess.call(cmd)  # nosec



import os
import tempfile

from detect_secrets import SecretsCollection
from detect_secrets.settings import transient_settings
from gibberish_detector import detector

# Secrets detection with detect-secrets tool


filters = [
    # some filters from [original list](https://github.com/Yelp/detect-secrets/blob/master/docs/filters.md#built-in-filters)
    # were removed based on their targets
    {"path": "detect_secrets.filters.heuristic.is_potential_uuid"},
    {"path": "detect_secrets.filters.heuristic.is_likely_id_string"},
    {"path": "detect_secrets.filters.heuristic.is_templated_secret"},
    {"path": "detect_secrets.filters.heuristic.is_sequential_string"},
]
plugins = [  # ruff: noqa: ERA001
    {"name": "ArtifactoryDetector"},
    {"name": "AWSKeyDetector"},
    # the entropy detectors esp Base64 need the gibberish detector on top
    {"name": "Base64HighEntropyString"},
    {"name": "HexHighEntropyString"},
    {"name": "AzureStorageKeyDetector"},
    {"name": "CloudantDetector"},
    {"name": "DiscordBotTokenDetector"},
    {"name": "GitHubTokenDetector"},
    {"name": "IbmCloudIamDetector"},
    {"name": "IbmCosHmacDetector"},
    {"name": "JwtTokenDetector"},
    {"name": "MailchimpDetector"},
    {"name": "NpmDetector"},
    {"name": "SendGridDetector"},
    {"name": "SlackDetector"},
    {"name": "SoftlayerDetector"},
    {"name": "StripeDetector"},
    {"name": "TwilioKeyDetector"},
    # remove 3 plugins for keyword
    # {'name': 'BasicAuthDetector'},
    # {'name': 'KeywordDetector'},
    # {'name': 'PrivateKeyDetector'},
]


def is_gibberish(matched_str):
    """Checks to make sure the PII span is gibberish and not word like."""
    # pip install gibberish-detector
    # download the training corpora from https://raw.githubusercontent.com/domanchi/gibberish-detector/master/examples/big.txt
    # run gibberish-detector train big.txt > big.model to generate the model (it takes 3 seconds)
    Detector = detector.create_from_model("gibberish_data/big.model")
    return Detector.is_gibberish(matched_str.lower())


def is_hash(content, value):
    """Second check if the value is a hash (after gibberish detector)."""
    # get the line where value occurred
    try:
        content.index(value)
    except ValueError:
        # TODO: fix this issue happened one for JS in the stack-smol, file did contain value
        print("Value not found in content, why this happened?")
        return False
    lines = content[: content.index(value)].splitlines()
    target_line = lines[-1]
    if len(value) in [32, 40, 64]:
        # if "sha" or "md5" are in content:
        keywords = ["sha", "md5", "hash", "byte"]
        if any(x in target_line.lower() for x in keywords):
            return True
    return False


def file_has_hashes(content, coeff=0.02):
    """Checks if the file contains literals 'hash' or 'sha' for more than 2% nb_of_lines."""
    lines = content.splitlines()
    count_sha = 0
    count_hash = 0
    nlines = content.count("\n")
    threshold = int(coeff * nlines)
    for line in lines:
        count_sha += line.lower().count("sha")
        count_hash += line.lower().count("hash")
        if count_sha > threshold or count_hash > threshold:
            return True
    return False


def get_indexes(text, value):
    string = text
    indexes = []
    new_start = 0
    while True:
        try:
            start = string.index(value)
            indexes.append(new_start + start)
            new_start = new_start + start + len(value)
            string = text[new_start:]
        except ValueError:
            break
    indexes = [(x, x + len(value)) for x in indexes]
    return indexes


def detect_keys(content, suffix=".txt"):
    """Detect secret keys in content using detect-secrets tool
    Args:
        content (str): string containing the text to be analyzed.
        suffix (str): suffix of the file
    Returns:
        A list of dicts containing the tag type, the matched string, and the start and
    end indices of the match.
    """
    fp = tempfile.NamedTemporaryFile(suffix=suffix, delete=False, mode="w")
    fp.write(content)
    fp.close()
    secrets = SecretsCollection()
    with transient_settings(
        {"plugins_used": plugins, "filters_used": filters},
    ):
        secrets.scan_file(fp.name)
    os.unlink(fp.name)
    secrets_set = list(secrets.data.values())
    matches = []
    if secrets_set:
        for secret in secrets_set[0]:
            if not is_gibberish(secret.secret_value):
                continue
            if is_hash(content, secret.secret_value) or file_has_hashes(content):
                continue
            indexes = get_indexes(content, secret.secret_value)
            for start, end in indexes:
                matches.append(
                    {
                        "tag": "KEY",
                        "value": secret.secret_value,
                        "start": start,
                        "end": end,
                    },
                )
    return matches



"""This component that captions images using a model from the Hugging Face hub."""
import io
import logging
import typing as t

import numpy as np
import pandas as pd
import torch
from PIL import Image
from transformers import BatchEncoding, BlipForConditionalGeneration, BlipProcessor

from fondant.component import PandasTransformComponent

logger = logging.getLogger(__name__)


def process_image(image: bytes, *, processor: BlipProcessor, device: str) -> torch.Tensor:
    """
    Process the image to a tensor.

    Args:
        image: The input image as a byte string.
        processor: The processor object for transforming the image.
        device: The device to move the transformed image to.
    """
    def load(img: bytes) -> Image:
        """Load the bytestring as an image."""
        bytes_ = io.BytesIO(img)
        return Image.open(bytes_).convert("RGB")

    def transform(img: Image) -> BatchEncoding:
        """Transform the image to a tensor using a processor and move it to the specified device."""
        return processor(images=img, return_tensors="pt").to(device)

    return transform(load(image))["pixel_values"]


def caption_image_batch(
    image_batch: pd.DataFrame,
    *,
    model: BlipForConditionalGeneration,
    processor: BlipProcessor,
    max_new_tokens: int,
) -> pd.Series:
    """Caption a batch of images."""
    input_batch = torch.cat(image_batch.tolist())
    output_batch = model.generate(pixel_values=input_batch, max_new_tokens=max_new_tokens)
    captions_batch = processor.batch_decode(output_batch, skip_special_tokens=True)

    return pd.Series(captions_batch, index=image_batch.index)


class CaptionImagesComponent(PandasTransformComponent):
    """Component that captions images using a model from the Hugging Face hub."""

    def setup(self, *, model_id: str, batch_size: int, max_new_tokens: int) -> None:
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        logger.info(f"Device: {self.device}")

        self.processor = BlipProcessor.from_pretrained(model_id)
        self.model = BlipForConditionalGeneration.from_pretrained(model_id).to(self.device)

        self.batch_size = batch_size
        self.max_new_tokens = max_new_tokens

    def transform(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        images = dataframe["images"]["data"].apply(
            process_image,
            processor=self.processor,
            device=self.device,
        )

        results: t.List[pd.Series] = []
        for batch in np.split(images, np.arange(self.batch_size, len(images), self.batch_size)):
            if not batch.empty:
                results.append(
                    caption_image_batch(
                        batch,
                        model=self.model,
                        processor=self.processor,
                        max_new_tokens=self.max_new_tokens,
                    ).T,
                )

        return pd.concat(results).to_frame(name=("captions", "text"))


if __name__ == "__main__":
    component = CaptionImagesComponent.from_args()
    component.run()



"""This component that segments images using a model from the Hugging Face hub."""
import io
import logging
import typing as t

import numpy as np
import pandas as pd
import torch
from palette import palette
from PIL import Image
from transformers import AutoModelForSemanticSegmentation, BatchFeature, SegformerImageProcessor

from fondant.component import PandasTransformComponent

logger = logging.getLogger(__name__)


def convert_to_rgb(seg: np.array):
    """
    Converts a 2D segmentation to a RGB one which makes it possible to visualize it.

    Args:
        seg: 2D segmentation map as a NumPy array.

    Returns:
        color_seg: 3D segmentation map contain RGB values for each pixel.
    """
    color_seg = np.zeros(
        (seg.shape[0], seg.shape[1], 3), dtype=np.uint8,
    )  # height, width, 3

    for label, color in enumerate(palette):
        color_seg[seg == label, :] = color

    color_seg = color_seg.astype(np.uint8).tobytes()

    return color_seg


def process_image(image: bytes, *, processor: SegformerImageProcessor, device: str) -> torch.Tensor:
    """
    Process the image to a tensor.

    Args:
        image: The input image as a byte string.
        processor: The processor object for transforming the image.
        device: The device to move the transformed image to.
    """
    def load(img: bytes) -> Image:
        """Load the bytestring as an image."""
        bytes_ = io.BytesIO(img)
        return Image.open(bytes_).convert("RGB")

    def transform(img: Image) -> BatchFeature:
        """Transform the image to a tensor using a clip processor and move it to the specified
        device.
        """
        return processor(images=img, return_tensors="pt").to(device)

    return transform(load(image))["pixel_values"]


@torch.no_grad()
def segment_image_batch(image_batch: pd.DataFrame, *, model: AutoModelForSemanticSegmentation,
                        processor: SegformerImageProcessor) -> pd.Series:
    """Embed a batch of images."""
    input_batch = torch.cat(image_batch.tolist())
    output_batch = model(input_batch)
    post_processed_batch = processor.post_process_semantic_segmentation(
        output_batch,
    )
    segmentations_batch = [convert_to_rgb(seg.cpu().numpy()) for seg in post_processed_batch]
    return pd.Series(segmentations_batch, index=image_batch.index)


class SegmentImagesComponent(PandasTransformComponent):
    """Component that segments images using a model from the Hugging Face hub."""

    def setup(self, model_id: str, batch_size: int) -> None:
        """
        Args:
            model_id: id of the model on the Hugging Face hub
            batch_size: batch size to use.
        """
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        logger.info(f"Device: {self.device}")

        self.processor = SegformerImageProcessor.from_pretrained(model_id)
        self.model = AutoModelForSemanticSegmentation.from_pretrained(model_id).to(self.device)

        self.batch_size = batch_size

    def transform(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        images = dataframe["images"]["data"].apply(
            process_image,
            processor=self.processor,
            device=self.device,
        )

        results: t.List[pd.Series] = []
        for batch in np.split(images, np.arange(self.batch_size, len(images), self.batch_size)):
            if not batch.empty:
                results.append(
                    segment_image_batch(
                        batch,
                        model=self.model,
                        processor=self.processor,
                    ).T,
                )

        return pd.concat(results).to_frame(name=("segmentations", "data"))


if __name__ == "__main__":
    component = SegmentImagesComponent.from_args()
    component.run()



import ipaddress
import json
import random
import string

# List of random private IP addresses to use as replacements
REPLACEMENTS_IP = {
    "IPv4": [
        "172.16.31.10",
        "172.16.58.3",
        "172.16.17.32",
        "192.168.127.12",
        "192.168.3.11",
    ],
    "IPv6": [
        "fd00:c2b6:b24b:be67:2827:688d:e6a1:6a3b",
        "fd00:a516:7c1b:17cd:6d81:2137:bd2a:2c5b",
        "fc00:e968:6179::de52:7100",
        "fc00:db20:35b:7399::5",
        "fdf8:f53e:61e4::18",
    ],
}

POPULAR_DNS_SERVERS = [
    "8.8.8.8",
    "8.8.4.4",
    "1.1.1.1",
    "1.0.0.1",
    "76.76.19.19",
    "76.223.122.150",
    "9.9.9.9",
    "149.112.112.112",
    "208.67.222.222",
    "208.67.220.220",
    "8.26.56.26",
    "8.20.247.20",
    "94.140.14.14",
    "94.140.15.15",
]


def load_json(sample):
    try:
        return json.loads(sample)
    except ValueError:
        return []


def random_replacements(n=10):
    """Build dictionaries of random replacements for PII (key, email, IP address).

    Emails: replace with one of n [random string of 5 characters + @example.com]
    IP addresses: replace with one of n synthetic private IP addresses (IPv4 or IPv6)
    Keys: replace with one of n [sequence of 32 random characters/digits]

    TODO: add IPv6 and IPv4 separation
    """
    letters = string.ascii_lowercase
    lettters_digits = string.ascii_lowercase + string.digits
    emails = [
        "".join(random.choice(letters) for i in range(5)) + "@example.com"
        for i in range(n)
    ]
    keys = [
        "".join(random.choice(lettters_digits) for i in range(32)) for i in range(n)
    ]
    ip_addresses = REPLACEMENTS_IP
    return {"EMAIL": emails, "KEY": keys, "IP_ADDRESS": ip_addresses}


def replace_ip(value, replacements_dict):
    """Replace an IP address with a synthetic IP address of the same format."""
    try:
        ipaddress.IPv4Address(value)
        return random.choice(replacements_dict["IP_ADDRESS"]["IPv4"])
    except ValueError:
        try:
            ipaddress.IPv6Address(value)
            return random.choice(replacements_dict["IP_ADDRESS"]["IPv6"])
        except ValueError:
            # this doesn't happen if we already use ipaddress filter in the detection
            print("Invalid IP address")
            return value


def is_private_ip(ip):
    """Check if an IP address is allocated for private networks."""
    ip = ipaddress.ip_address(ip)
    return ip.is_private


def redact_pii_text(text, secrets, replacements, add_references=False):
    """Redact PII in a text
    Args:
        text (str): text to redact
        secrets (json): json string with the secrets to redact
        replacements (dict): dictionary of replacements for each PII type
        add_references (bool): whether to add references to the redacted text (delimiters to PII)
        for visualization
    Returns:
        text (str): new text with redacted secrets.
    """
    secrets = load_json(secrets)
    modified = False
    if secrets:
        secrets = sorted(secrets, key=lambda x: x["start"])
        # store the secrets that were replaced here with their replacements
        replaced_secrets = {}
        subparts = []
        references = []
        step = 0
        last_text = text
        for secret in secrets:
            # skip secret if it's an IP address for private networks or popular DNS servers
            if secret["tag"] == "IP_ADDRESS":  # ruff: noqa: SIM102
                # if secret value in popular DNS servers, skip it
                if is_private_ip(secret["value"]) or (
                    secret["value"] in POPULAR_DNS_SERVERS
                ):
                    continue
            modified = True
            subtext = text[step : secret["start"]]
            subpart = subtext if subtext else " "
            subparts.append(subpart)
            # if secret is already in replaced_secrets, use the same replacement
            if secret["value"] in replaced_secrets:
                replacement = replaced_secrets[secret["value"]]
            else:
                if secret["tag"] == "IP_ADDRESS":
                    replacement = replace_ip(secret["value"], replacements)
                else:
                    replacement = random.choice(replacements[secret["tag"]])
                replaced_secrets[secret["value"]] = replacement
            subparts.append(replacement)
            replaced_secrets[secret["value"]] = replacement
            if add_references:
                references.append(subpart)
                references.append(f"PI:{secret['tag']}:{replacement}END_PI")
            last_text = text[secret["end"] :]
            step = secret["end"]
        # if subparts are not empty join them (it can be empty when all secrets were skipped)
        new_text = "".join(subparts) + last_text if subparts else last_text
        if add_references:
            references = "".join(references) + last_text if references else ""
    else:
        new_text = text
        references = ""
    return (
        (new_text, references, modified) if add_references else (new_text, modified)
    )


def redact_pii(text, secrets, has_secrets, replacements):
    """Anonymize PII in an example from a dataset.

    Slightly modified to only return the potentially updated text.
    """
    if has_secrets:
        new_text, _ = redact_pii_text(text, secrets, replacements)
        return new_text

    return text



"""This component filters images of the dataset based on image size
(minimum image size and maximumn aspect ratio).
"""
import logging

import numpy as np
import pandas as pd

from fondant.component import PandasTransformComponent

logger = logging.getLogger(__name__)


class FilterImageResolutionComponent(PandasTransformComponent):
    """Component that filters images based on height and width."""

    def setup(self, *, min_image_dim: int, max_aspect_ratio: float) -> None:
        self.min_image_dim = min_image_dim
        self.max_aspect_ratio = max_aspect_ratio

    def transform(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """
        Args:
            dataframe: Pandas dataframe
            min_image_dim: minimum image dimension.
            min_aspect_ratio: minimum aspect ratio.

        Returns:
            filtered Pandas dataframe
        """
        width = dataframe["image"]["width"]
        height = dataframe["image"]["height"]
        min_image_dim = np.minimum(width, height)
        max_image_dim = np.maximum(width, height)
        aspect_ratio = max_image_dim / min_image_dim
        mask = (min_image_dim >= self.min_image_dim) & (aspect_ratio <= self.max_aspect_ratio)

        return dataframe[mask]


if __name__ == "__main__":
    component = FilterImageResolutionComponent.from_args()
    component.run()



"""This component retrieves image URLs from LAION-5B based on a set of CLIP embeddings."""
import asyncio
import concurrent.futures
import functools
import logging
import typing as t

import pandas as pd
from clip_client import ClipClient, Modality

from fondant.component import PandasTransformComponent

logger = logging.getLogger(__name__)


class LAIONRetrievalComponent(PandasTransformComponent):
    """Component that retrieves image URLs from LAION-5B based on a set of CLIP embeddings."""

    def setup(
            self,
            *,
            num_images: int,
            aesthetic_score: int,
            aesthetic_weight: float,
    ) -> None:
        """

        Args:
            num_images: number of images to retrieve for each prompt
            aesthetic_score: ranking score for aesthetic embedding, higher is prettier,
                between 0 and 9.
            aesthetic_weight: weight of the aesthetic embedding to add to the query,
                between 0 and 1.
        """
        self.client = ClipClient(
            url="https://knn.laion.ai/knn-service",
            indice_name="laion5B-L-14",
            num_images=num_images,
            aesthetic_score=aesthetic_score,
            aesthetic_weight=aesthetic_weight,
            modality=Modality.IMAGE,
        )

    def transform(
            self,
            dataframe: pd.DataFrame,
    ) -> pd.DataFrame:
        """Asynchronously retrieve image URLs and ids based on prompts in the provided dataframe."""
        results: t.List[t.Tuple[str]] = []
        loop = asyncio.new_event_loop()

        async def async_query():
            with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
                futures = [
                    loop.run_in_executor(
                        executor,
                        functools.partial(self.client.query, embedding_input=embedding.tolist()),
                    )
                    for embedding in dataframe["embeddings"]["data"]
                ]
                for response in await asyncio.gather(*futures):
                    results.extend(response)

        loop.run_until_complete(async_query())

        results_df = pd.DataFrame(results)[["id", "url"]]
        results_df = results_df.set_index("id")
        results_df.columns = [["images"], ["url"]]

        return results_df


if __name__ == "__main__":
    component = LAIONRetrievalComponent.from_args()
    component.run()
